%%%-------------------------------------------------------------------
%%% @author Rick Payne <rickp@rossfell.co.uk>
%%% @copyright (C) 2014, Alistair Woodman, California USA <awoodman@netdef.org>
%%% @doc
%%%
%%% Handle an IS-IS level for a given interface.
%%%
%%% This file is part of AutoISIS.
%%%
%%% License:
%%% This code is licensed to you under the Apache License, Version 2.0
%%% (the "License"); you may not use this file except in compliance with
%%% the License. You may obtain a copy of the License at
%%% 
%%%   http://www.apache.org/licenses/LICENSE-2.0
%%% 
%%% Unless required by applicable law or agreed to in writing,
%%% software distributed under the License is distributed on an
%%% "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
%%% KIND, either express or implied.  See the License for the
%%% specific language governing permissions and limitations
%%% under the License.
%%%
%%% @end
%%% Created :  7 Feb 2014 by Rick Payne <rickp@rossfell.co.uk>
%%%-------------------------------------------------------------------
-module(isis_config).

-behaviour(gen_server).

%% API
-export([start_link/0,
	 set/2,
	 get/1, get_item/2, get_item/3,
	 get_all_items/2]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
	 terminate/2, code_change/3]).

-define(SERVER, ?MODULE).
-define(CONFIG_ETS, isis_config).

-record(config_item, {
	  key = [],    %% List of items in the form [{interface, "eth0"}, {level, level_1}]
	  value        %% Individual item.
	 }).

-record(state, {}).

%%%===================================================================
%%% API
%%%===================================================================

%%--------------------------------------------------------------------
%% @doc
%% Starts the server
%%
%% @spec start_link() -> {ok, Pid} | ignore | {error, Error}
%% @end
%%--------------------------------------------------------------------
start_link() ->
    gen_server:start_link({local, ?SERVER}, ?MODULE, [], []).

set(Key, Value) ->
    gen_server:call(?MODULE, {set, Key, Value}).

get(Key) ->
    lists:map(fun(#config_item{value = V}) -> V end,
	      ets:lookup(?CONFIG_ETS, Key)).

get_item(Key, Item) ->
    proplists:get_value(Item, ?MODULE:get(Key)).
get_item(Key, Item, Default) ->
    proplists:get_value(Item, ?MODULE:get(Key), Default).

get_all_items(Key, Item) ->
    proplists:get_all_values(Item, ?MODULE:get(Key)).

%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Initializes the server
%%
%% @spec init(Args) -> {ok, State} |
%%                     {ok, State, Timeout} |
%%                     ignore |
%%                     {stop, Reason}
%% @end
%%--------------------------------------------------------------------
init([]) ->
    ets:new(?CONFIG_ETS, [named_table, bag,
			  {keypos, #config_item.key}]),
    get_initial_config(),
    {ok, #state{}}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling call messages
%%
%% @spec handle_call(Request, From, State) ->
%%                                   {reply, Reply, State} |
%%                                   {reply, Reply, State, Timeout} |
%%                                   {noreply, State} |
%%                                   {noreply, State, Timeout} |
%%                                   {stop, Reason, Reply, State} |
%%                                   {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_call({set, Key, Value}, _From, State)->
    insert_items(Key, Value),
    {reply, ok, State};
handle_call(_Request, _From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling cast messages
%%
%% @spec handle_cast(Msg, State) -> {noreply, State} |
%%                                  {noreply, State, Timeout} |
%%                                  {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_cast(_Msg, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling all non call/cast messages
%%
%% @spec handle_info(Info, State) -> {noreply, State} |
%%                                   {noreply, State, Timeout} |
%%                                   {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_info(_Info, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any
%% necessary cleaning up. When it returns, the gen_server terminates
%% with Reason. The return value is ignored.
%%
%% @spec terminate(Reason, State) -> void()
%% @end
%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    ok.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Convert process state when code is changed
%%
%% @spec code_change(OldVsn, State, Extra) -> {ok, NewState}
%% @end
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%%===================================================================
%%% Internal functions
%%%===================================================================
insert_items(Key, Value) when is_list(Value) ->
    Keys = proplists:get_keys(Value),
    lists:map(fun(#config_item{key = _, value = {A, _}} = I) ->
		      case lists:member(A, Keys) of
			  true -> ets:delete_object(?CONFIG_ETS, I);
			  false -> ok
		      end
	      end,
	      ets:lookup(?CONFIG_ETS, Key)),
    Vs =
	lists:map(
	  fun(V) -> #config_item{key = Key, value = V} end,
	  Value),
    ets:insert(?CONFIG_ETS, Vs);
insert_items(Key, Value) ->
    insert_items(Key, [Value]).
				

get_initial_config() ->
    case lists:member(isis_system, registered()) of
	true ->
	    lists:map(
	      fun({K, V}) -> insert_items(K, V) end,
	      isis_system:get_state(isis_config));
	_ -> get_initial_config()
    end.
